@@ -1,36 +1,43 @@
 /*****************************************************************************
- * Secure Locate v1.3                                                        *
- * October 9, 1998                                                           *
+ * Secure Locate v1.4                                                        *
+ * October 24, 1998                                                          *
  * Programmed by: Kevin Lindsay                                              *
  * Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.          *
  *                                                                           *
  * NetNation Communications:  http://www.netnation.com/                      *
  * Secure Locate: ftp://ftp.mkintraweb.com/pub/linux/slocate/                *
  *                                                                           *
+ * Patches:  Sean Mcnulty <lazy@ccf.rutgers.edu>                             *
+ *                * Fixed a bug which caused the decode function to fail.    *
+ *                                                                           *
  * Report any Bugs to: klindsay@mkintraweb.com                               *
  *                                                                           *
  * Even though Secure Locate was written from scratch, some recognition goes *
  * to James A. Woods <jwoods@adobe.com> for writing the original GNU locate. *
  *****************************************************************************/
 
-/* Secure Locate -- search database for filenames that match patterns without
- *                  showing files that the user using slocate does not have
- *                  access to.
- *    Copyright (C) 1994 Free Software Foundation, Inc.
- * 
- *    This program is free software; you can redistribute it and/or modify
- *    it under the terms of the GNU General Public License as published by
- *    the Free Software Foundation; either version 2, or (at your option)
- *    any later version.
- * 
- *    This program is distributed in the hope that it will be useful,
- *    but WITHOUT ANY WARRANTY; without even the implied warranty of
- *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *    GNU General Public License for more details.
- * 
- *    You should have received a copy of the GNU General Public License
- *    along with this program; if not, write to the Free Software
- *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+/*****************************************************************************
+ *                                                                            
+ * Secure Locate -- search database for filenames that match patterns without 
+ *                  showing files that the user using slocate does not have   
+ *                  access to.                                                
+ *    Copyright (C) 1994 Free Software Foundation, Inc.                       
+ *                                                                            
+ *    This program is free software; you can redistribute it and/or modify    
+ *    it under the terms of the GNU General Public License as published by    
+ *    the Free Software Foundation; either version 2, or (at your option)     
+ *    any later version.                                                      
+ *                                                                            
+ *    This program is distributed in the hope that it will be useful,         
+ *    but WITHOUT ANY WARRANTY; without even the implied warranty of          
+ *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           
+ *    GNU General Public License for more details.                            
+ *                                                                            
+ *    You should have received a copy of the GNU General Public License       
+ *    along with this program; if not, write to the Free Software             
+ *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               
+ *                                                                            
+ *****************************************************************************/
 
 #include <stdlib.h>
 #include <unistd.h>
@@ -43,24 +50,34 @@
 #include <dirent.h>
 #include <netinet/in.h>
 #include <grp.h>
+#include <ctype.h>
+#include <fnmatch.h>
 
 #include "statmalloc.h"
 #include "link.h"
 
 /* GLOBALS */
 
-#define VERSION "Secure Locate v1.3 - Released October 14, 1998\n"
-#define SLOCATEDB "/var/lib/slocate/slocate.db"
-#define TMPSLOCATEDB "/var/lib/slocate/slocate.tmp"
-#define SLOCATEDB_DIR "/var/lib/slocate/"
+#define VERSION "Secure Locate v1.4 - Released October 24, 1998\n"
 #define GRPFILE "/etc/group"
 #define SLOC_ESC 0x80
 #define SLOC_GRP "slocate"
 #define SLOC_UID 0
 #define MIN_BLK 64
+/* Warn if a database is older than this.  8 days allows for a weekly
+ *    update that takes up to a day to perform.  */
+#define WARN_SECONDS (60 * 60 * 24 * 8)
 
+/* Printable version of WARN_SECONDS.  */
+#define WARN_MESSAGE "8 days"
+
+char *SLOCATEDB = "/var/lib/slocate/slocate.db";
+char *TMPSLOCATEDB = "/var/lib/slocate/slocate.tmp";
+char *SLOCATEDB_DIR = "/var/lib/slocate/";
 char *EXCLUDE_DIR;
 int EXCLUDE=0;
+int VERBOSE=0;
+int QUIET=0;
 char *progname;
 char *tmp_path=NULL;
 char *cur_dir=NULL;
@@ -77,22 +94,32 @@ unsigned short SLOC_GID=65534; /* Seems to be the default nobody gid */
 void
 usage() 
 {
-   printf("Secure Locate v1.3\n"
-          "Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.\n"
-          "\n"
-          "usage: %s [options] <search string>\n\n"
+   printf("%s\n"
+          "Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.\n\n"
+          "search usage:   %s [-q] [-d path] [--database=path] <search string>\n"
+          "database usage: %s [-qv] [-e dir1,dir2,...] [-l level] <[-U path] [-u]>\n"
+          "general usage:  %s [-Vh] [--version] [--help]\n\n"
           "   Options:\n"
           "   -u                 - Create slocate database starting at path /.\n"
           "   -U <dir>           - Create slocate database starting at path <dir>.\n"          
-          "   -e <dir1,dir2,...> - Exclude directories from slocate database.\n"
-          "   -l <[01]>          - Security level. \n"
+          "   -e <dir1,dir2,...> - Exclude directories from the slocate database.\n"
+          "   -l <level>         - Security level. \n"
           "                           0 turns security checks off. This will make\n"
           "                             searchs faster.\n"
           "                           1 turns security checks on. This is the default.\n"
+          "   -q                 - Quite mode.  Error messages are suppressed.\n"
+          "   -d <path>\n"
+          "   --database=path    - Specfies path of database to search in.\n"
+          "   -h\n"
+          "   --help             - Display this help.\n"
+          "   -v\n"
+          "   --verbose          - Verbose mode. Display files when creating database.\n"
+          "   -V\n"
+          "   --version          - Display version.\n"
           "\n"
           "Author: Kevin Lindsay\n"
           "Bugs: klindsay@mkintraweb.com\n"
-          "\n",progname);
+          "\n",VERSION,progname,progname,progname);
    
    exit(0);
 }
@@ -102,7 +129,8 @@ int
 MYfchdir(int fd, char *filepath, char *func)
 {   
    if (fchdir(fd) == -1) {
-      fprintf(stderr,"%s: %s: fchdir: %s: %s\n",progname,func,filepath,strerror(errno));
+      if (!QUIET)
+          fprintf(stderr,"%s: %s: fchdir: %s: %s\n",progname,func,filepath,strerror(errno));
       return(0);
    }
    return(1);
@@ -113,7 +141,8 @@ int
 MYchdir(char *filepath, char *func)
 {
    if (chdir(filepath) == -1) {
-      fprintf(stderr,"%s: %s: chdir: %s: %s\n",progname,func,filepath,strerror(errno));
+      if (!QUIET)      
+          fprintf(stderr,"%s: %s: chdir: %s: %s\n",progname,func,filepath,strerror(errno));
       return(0);
    }
    return(1);
@@ -135,20 +164,95 @@ get_short (FILE *fp)
    return (x << 8) | (fgetc (fp) & 0xff);
 }
 
+/* Parse Path */
+void
+parse_path(char *path)
+{
+   int ptrlen;
+   int grow;
+   struct stat statres;
+   char *ptr;
+
+   ptrlen = strlen(path);   
+   if (stat(path,&statres) == -1) {
+      if (!QUIET)
+          fprintf(stderr,"%s: error accessing path: %s : %s\n",progname,path,strerror(errno));
+      exit(1);
+   }
+   
+   if (S_ISDIR(statres.st_mode)) {
+      if (!QUIET)
+          fprintf(stderr,"%s: %s is a directory\n",progname,path);
+      exit(1);
+   }
+
+   grow=ptrlen+1;
+   SLOCATEDB = malloc(grow);
+   SLOCATEDB[0] = '\0';
+   strcat(SLOCATEDB,path);
+   
+   if ((ptr = strrchr(path,'/')) != NULL) {
+      ptr++;
+      SLOCATEDB_DIR = malloc((ptr-path)+1);
+      SLOCATEDB_DIR[0] = '\0';
+      strncat(SLOCATEDB_DIR,path,ptr-path);
+   } else {
+      SLOCATEDB_DIR = malloc(8);
+      strcat(SLOCATEDB_DIR,".\0");
+   }
+
+   TMPSLOCATEDB = malloc(strlen(SLOCATEDB)+5);
+   TMPSLOCATEDB[0] = '\0';
+   strcat(TMPSLOCATEDB,SLOCATEDB);
+   strcat(TMPSLOCATEDB,".tmp");
+
+}
+
+/* Parse Dash */
+
+void
+parse_dash(char *option)
+{
+   char *ptr;
+   
+   for (ptr = option;*ptr != 0 && *ptr != '=' ; ptr++)
+       *ptr = toupper(*ptr);
+
+   if (!strcmp(option,"HELP"))
+       usage();
+       
+   if (!strcmp(option,"VERSION")) {
+      printf("%s",VERSION);
+      exit(0);
+   }
+   
+   if (!strcmp(option,"VERBOSE"))
+       VERBOSE=1;
+   
+   *ptr = '\0';
+   ptr++;
+   if (!strcmp(option,"DATABASE")) {
+      parse_path(ptr);
+   }
+   
+}
+
+
 /* Make sure that the DB Directory exists */
 
 int
 check_dir()
 {
-	   int res=1;
-	   struct stat tmpstat;
-	
-	   if (lstat(SLOCATEDB_DIR,&tmpstat) == -1) {
-			      fprintf(stderr,"%s: error accessing DB Directory: %s : %s\n",progname,SLOCATEDB_DIR,strerror(errno));
-			      res=0;
-		}
-	
-	   return(res);
+   int res=1;
+   struct stat tmpstat;
+   
+   if (lstat(SLOCATEDB_DIR,&tmpstat) == -1) {
+      if (!QUIET)
+          fprintf(stderr,"%s: error accessing DB Directory: %s : %s\n",progname,SLOCATEDB_DIR,strerror(errno));
+      res=0;
+   }
+   
+   return(res);
 }
 
 /* Get the GID for group slocate */
@@ -157,15 +261,17 @@ unsigned short
 get_gid()
 {
    unsigned short GGID=SLOC_GID;
-	struct group *grpres;
-	
-	if ((grpres = getgrnam(SLOC_GRP)) == NULL) {		
-      fprintf(stderr,"%s: Could not find the group: %s in the /etc/group file.\n",progname,SLOC_GRP);
-      fprintf(stderr,"This is a result of the group missing or a corrupted group file.\n");
+   struct group *grpres;
+   
+   if ((grpres = getgrnam(SLOC_GRP)) == NULL) {
+      if (!QUIET) {
+         fprintf(stderr,"%s: Could not find the group: %s in the /etc/group file.\n",progname,SLOC_GRP);
+         fprintf(stderr,"This is a result of the group missing or a corrupted group file.\n");
+      }
       exit(1);
-	}
+   }
 
-	GGID = grpres->gr_gid;
+   GGID = grpres->gr_gid;
    
    return GGID;
 }
@@ -198,16 +304,23 @@ parse_exclude(char *estr)
       }
 
       grow += (ptr2-ptr1)+2;
-      excludestr = realloc(excludestr,grow);		
+      if (*(ptr2-1) == '/') {
+         grow--;
+         ptr2--;
+      }
+      
+      excludestr = realloc(excludestr,grow);
       strcat(excludestr,"*");
       strncat(excludestr,ptr1,ptr2-ptr1);
       strcat(excludestr,"*");
+      
+      if (*ptr2 == '/')
+          ptr2++;
+
       if (ptr2[0] == ',')
           ptr1 = ptr2+1;
    }
-  
-	printf("E: %s\n",excludestr);
-	exit(1);
+   
    return (excludestr);
 }
 
@@ -244,13 +357,15 @@ load_dir(int dir_fd, dir_item *listptr)
    if (MYfchdir(dir_fd,listptr->name,"LD_1") == -1) return(NULL);
    
    if ((recdir = opendir(".")) == NULL) {
-      fprintf(stderr,"%s: Opendir: %s: %s\n",progname,listptr->name,strerror(errno));
+      if (!QUIET)
+          fprintf(stderr,"%s: Opendir: %s: %s\n",progname,listptr->name,strerror(errno));
       return(NULL);
    }
 
    /* Open Current Directory 
    if ((filename = readdir(recdir)) == NULL) {
-      fprintf(stderr,"%s: Readdir Error: Load List: %s: %s\n",progname,listptr->name,strerror(errno));
+      if (!QUIET)
+          fprintf(stderr,"%s: Readdir Error: Load List: %s: %s\n",progname,listptr->name,strerror(errno));
       closedir(recdir);
       return(NULL);
    }
@@ -297,7 +412,8 @@ load_dir(int dir_fd, dir_item *listptr)
       listptr->name_len = strlen(listptr->name);      
 
       if (lstat(filename->d_name,&statres) == -1) {
-         fprintf(stderr,"%s: lstat: %s: %s\n",progname,filename->d_name,strerror(errno));
+         if (!QUIET)
+             fprintf(stderr,"%s: lstat: %s: %s\n",progname,filename->d_name,strerror(errno));
       }
       
       if (S_ISDIR(statres.st_mode))
@@ -391,6 +507,9 @@ frcode(char *dir_path, char *filename, struct stat statres)
    strcat(cur_path,dir_path);
    strcat(cur_path,filename);
    
+   if (VERBOSE)
+       fprintf(stdout,"%s\n",cur_path);
+   
    if (tmp_path != NULL) {
       ptr1=cur_path;
       ptr2=tmp_path;
@@ -428,8 +547,8 @@ frcode(char *dir_path, char *filename, struct stat statres)
       putc (SLOC_ESC, fd);
       put_short (fr_num, fd);
    } else
-		 putc(fr_num,fd);
-	
+       putc(fr_num,fd);
+   
    fputs(new_path,fd);
    putc('\0',fd);
 
@@ -463,7 +582,8 @@ create_db(char *dirstr)
    FILE *fd;
 
    if (UID != SLOC_UID) {
-      fprintf(stderr,"%s: You are not authorized to create an slocate database!\n",progname);
+      if (!QUIET)
+          fprintf(stderr,"%s: You are not authorized to create an slocate database!\n",progname);
       return(0);
    }
    
@@ -474,7 +594,8 @@ create_db(char *dirstr)
    if (dirstr) {
       /* Make sure there are no problems accessing the source directory */
       if (lstat(dirstr,&statres) == -1) {
-         fprintf(stderr,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
+         if (!QUIET)
+             fprintf(stderr,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
          return(-1);
       }
    }
@@ -503,7 +624,8 @@ create_db(char *dirstr)
    
    /* Open Source Directory */
    if ((dir_fd = open(dirstr,O_RDONLY)) == -1) {
-      fprintf(stderr,"%s: open: %s: %s\n",progname,dirstr,strerror(errno));
+      if (!QUIET)
+          fprintf(stderr,"%s: open: %s: %s\n",progname,dirstr,strerror(errno));
       return(-1);
    }
    
@@ -578,7 +700,8 @@ create_db(char *dirstr)
             close(dir_fd);
             
             if ((dir_fd = open(".",O_RDONLY)) == -1) {
-               fprintf(stderr,"%s: open: CDB_3: %s\n",progname,strerror(errno));
+               if (!QUIET)
+                   fprintf(stderr,"%s: open: CDB_3: %s\n",progname,strerror(errno));
                return(1);
             }
             
@@ -600,7 +723,8 @@ create_db(char *dirstr)
             close(dir_fd);
             
             if ((dir_fd = open(".",O_RDONLY)) == -1) {
-               fprintf(stderr,"%s: open: CDB_5: %s\n",progname,strerror(errno));
+               if (!QUIET)
+                   fprintf(stderr,"%s: open: CDB_5: %s\n",progname,strerror(errno));
                return(1);
             }
          }
@@ -622,77 +746,116 @@ decode_db(char *str)
 {
    int res = 1;
    FILE *fd;
+   DIR *tmpfd;
    short code_num;
-	int pathlen=0;
+   int pathlen=0;
    register char ch;
    int grow=0;
    int jump=0;
-	int first=1;
-	char *codedpath=NULL;
-	char *ptr;
-   int printit;
-   struct stat statres;	
-  
+   int first=1;
+   char *codedpath=NULL;
+   char *tmpcodedpath;    /* to fix around realloc() problem */
+   char *ptr;
+   int printit=0;
+   int globflag=0;
+   char *globptr1;
+   struct stat statres;
+   time_t now;
+   char *chk1;
+   char tmpch;
+   
    if ((fd = fopen(SLOCATEDB,"r")) == NULL) {
-      fprintf(stderr,"%s: decode_db(): open: (%d) %s\n",progname,errno,strerror(errno));
+      if (!QUIET)
+          fprintf(stderr,"%s: decode_db(): open: (%d) %s\n",progname,errno,strerror(errno));
       return(0);
    }
 
+   lstat(SLOCATEDB,&statres);
+
+   time(&now);
+   if (now - statres.st_mtime > WARN_SECONDS)
+   {
+      fprintf(stdout,"%s: warning: database %s' is more than %s old\n",progname,SLOCATEDB,WARN_MESSAGE);
+   }   
+   
    slevel = getc(fd);
-	
-	codedpath = malloc(MIN_BLK);
-	codedpath[0] = '\0';
+   
+   codedpath = malloc(MIN_BLK);
+   codedpath[0] = '\0';
    ptr=codedpath;
+   
+   if ((globptr1 = strchr(str,'*')) != NULL ||
+       (globptr1 = strchr(str,'?')) != NULL ||
+       ((globptr1 = strchr(str,'[')) != NULL && strchr(str,']') != NULL))
+       globflag = 1;
+   
    while ((code_num = getc(fd)) != EOF) {
-		
+      
       if (code_num == SLOC_ESC) {
          code_num = get_short(fd);
       } else if (code_num > 127)
-				 code_num = code_num - 256;
-		
+             code_num = code_num - 256;
+
       jump=0;
-		
-		if (code_num < 0)
-			grow += code_num;
-		
-		ptr+=code_num;
-		
-		pathlen = ptr - codedpath;
+      
+      if (code_num < 0)
+         grow += code_num;
+      
+      ptr+=code_num;
+      
+      pathlen = ptr - codedpath;
 
       while(!jump) {
          ch = getc(fd);
          grow++;
-			pathlen++;
-			if (grow == 64) {
-				codedpath = realloc(codedpath,pathlen+MIN_BLK);
-				grow=0;
-			}
-         codedpath[pathlen-1] = ch;
-	
-         if (ch == '\0')
-				jump=1;
-      }
-		
-		if (first) {
-			ptr=ptr+strlen(codedpath);
-			first=0;
-		}
-		
-		pathlen--;
+         pathlen++;
+         if (grow == 64) {
 
-      if (strstr(codedpath,str) != NULL) {
+            tmpcodedpath = realloc(codedpath,pathlen+MIN_BLK);
+            ptr += (tmpcodedpath - codedpath);
+            codedpath = tmpcodedpath;
+            
+            grow=0;
+         }
+         codedpath[pathlen-1] = ch;
+   
+         if (ch == '\0')
+            jump=1;
+      }
+      
+      if (first) {
+         ptr=ptr+strlen(codedpath);
+         first=0;
+      }
+      
+      pathlen--;            
+      
+      printit=0;
+      
+      if ((globflag && !fnmatch(str,codedpath,0)) || strstr(codedpath,str) != NULL) {
          if (slevel == '1') {
-				
-            if (UID == 0 || lstat(codedpath,&statres) == 0)
-					 printit=1;
-				else
-					 printit=0;
+            if (UID == 0 || lstat(codedpath,&statres) == 0) {
+               printit=1;
+               chk1 = codedpath+1;
+               while ((chk1 = strchr(chk1,'/')) != NULL) {
+                  chk1++;
+                  tmpch=*chk1;
+                  *chk1='\0';
+
+                  if ((tmpfd = opendir(codedpath)) == NULL)
+                      printit=0;
+                  else
+                      closedir(tmpfd);
+                  
+                  *chk1=tmpch;
+               }
+            }
          } else
              printit=1;
-         
-         if (printit)
-             printf("%s\n",codedpath);
       }
+      
+      if (printit)
+         printf("%s\n",codedpath);
    }      
    
    fclose(fd);
@@ -715,30 +878,40 @@ main(int args, char **argv)
    char *spec_dir=NULL;
    int optlen;
 
-   progname = ((p = rindex(argv[0],'/')) ? p+1 : *argv);   
+   progname = ((p = strrchr(argv[0],'/')) ? p+1 : *argv);   
    
    if (args < 2) 
        usage();
 
-	if (!check_dir())
-		 return(1);	
-	
-   UID = getuid();        
-   GID = getgid();      
+   if (!check_dir())
+       return(1);   
    
-   while ((ch = getopt(args,argv,"VuU:e:l:")) != EOF) {
+   UID = getuid();        
+   GID = getgid();
+   
+   while ((ch = getopt(args,argv,"VvuhqU:e:l:d:-:")) != EOF) {
       switch(ch) {
+      case 'h':
+         usage();
+         break;
+      case 'q':
+         QUIET=1;
+         break;
       case 'V':
          printf("%s",VERSION);
          exit(1);
          break;
+      case 'v':
+         VERBOSE=1;
+         break;
       case 'e':
          EXCLUDE_DIR = parse_exclude(optarg);
          break;
       case 'l':
          slevel = optarg[0];
          if (slevel != '0' && slevel != '1') {
-            fprintf(stderr,"%s: ERROR: Security level must be 0 or 1.\n",progname);
+            if (!QUIET)
+                fprintf(stderr,"%s: ERROR: Security level must be 0 or 1.\n",progname);
             exit(1);
          }
          break;
@@ -752,6 +925,12 @@ main(int args, char **argv)
          memmove(spec_dir,optarg,optlen);
          spec_dir[optlen] = '\0';         
          break;
+      case 'd':
+         parse_path(optarg);
+         break;
+      case '-':
+         parse_dash(optarg);
+         break;
       default:
          return(1);
          break;
@@ -767,7 +946,7 @@ main(int args, char **argv)
    else if (ROOTDIR)
       res = create_db((char *)NULL);
    else if (argv[optind]) {
-      res = decode_db(argv[1]);
+      res = decode_db(argv[optind]);
    } else
        usage();
 
